@@ -233,6 +233,7 @@ parse(char **data, /* I/O - Data to process */
   int           linecount = 1;   /* Count the lines for error messages */
   int           nestinglevel = 0;/* How many XML environments are nested
 				    at the point where we are */
+  int           inxmlheader = 1;
   int           intag = 0;
   int           incomment = 0;
   int           tagnamefound = 0;
@@ -362,7 +363,7 @@ parse(char **data, /* I/O - Data to process */
     case '<': /* open angle bracket */
       if (!inquotes) {
 	if (intag) {
-	  if (!incomment) {
+	  if (!incomment && !inxmlheader) {
 	    /* Unless a tag is a comment, angle brackets cannot appear inside
 	       the tag. */
 	    fprintf(stderr, "XML error: Nested angle brackets in %s, line %d!\n",
@@ -378,8 +379,9 @@ parse(char **data, /* I/O - Data to process */
 	  intag = 1;
 	  if (scan + 3 < *data + datalength) {
 	    if ((*(scan + 1) == '!') && (*(scan + 2) == '-') &&
-		(*(scan + 3) == '-')) { 
+		(*(scan + 3) == '-')) {
 	      incomment = 1;
+	      tagtype = 0;
 	      if (debug) fprintf(stderr, "    Start of a comment\n");
 	    }
 	  }
@@ -433,6 +435,37 @@ parse(char **data, /* I/O - Data to process */
 		    inid = nestinglevel + 1;
 		  } else if (strcmp(currtagname, "printer") == 0) {
 		    inprinter = nestinglevel + 1;
+		    if (tagtype == 1) {
+		      /* XML body of the file is starting here */
+		      inxmlheader = 0;
+		      nestinglevel = 1;
+		      /* Remove the whole header of the XML file */
+		      if (debug) 
+			fprintf(stderr,
+				"    Removing XML file header\n");
+		      memmove(*data, lasttag, 
+			      *data + datalength + 1 - lasttag);
+		      datalength -= lasttag - *data;
+		      scan -= lasttag - *data;
+		      tagwordstart -= lasttag - *data;
+		      lasttag = *data;
+		      lasttagend = NULL;
+		    }
+		  } else if (strcmp(currtagname, "lang") == 0) {
+		    inlang = nestinglevel + 1;
+		  } else if (strcmp(currtagname, "postscript") == 0) {
+		    inpostscript = nestinglevel + 1;
+		    if (inlang) {
+		      if (tagtype == 1) {
+			if (debug)
+			  fprintf(stderr, 
+				  "    Resetting Driver/PPD.\n");
+			cid[0] = '\0';
+			cppd[0] = '\0';
+		      }
+		    }
+		  } else if (strcmp(currtagname, "ppd") == 0) {
+		    inppd = nestinglevel + 1;
 		  }
 		} else if (operation == 1) { /* Driver XML file */
 		  if (strcmp(currtagname, "printer") == 0) {
@@ -462,6 +495,22 @@ parse(char **data, /* I/O - Data to process */
 		    }
 		  } else if (strcmp(currtagname, "driver") == 0) {
 		    indriver = nestinglevel + 1;
+		    if (tagtype == 1) {
+		      /* XML body of the file is starting here */
+		      inxmlheader = 0;
+		      nestinglevel = 1;
+		      /* Remove the whole header of the XML file */
+		      if (debug) 
+			fprintf(stderr,
+				"    Removing XML file header\n");
+		      memmove(*data, lasttag, 
+			      *data + datalength + 1 - lasttag);
+		      datalength -= lasttag - *data;
+		      scan -= lasttag - *data;
+		      tagwordstart -= lasttag - *data;
+		      lasttag = *data;
+		      lasttagend = NULL;
+		    }
 		  } 
 		} else if (operation == 2) { /* Option XML file */
 		  if (strcmp(currtagname, "make") == 0) {
@@ -541,7 +590,23 @@ parse(char **data, /* I/O - Data to process */
 		    /* Mark up to the end of the tag before, to insert the
 		       definition of the default option setting */
 		    if (tagtype == -1) lastoption = (char*)lasttagend + 1;
-		    if (tagtype == 1) argdefault[0] = '\0';
+		    if (tagtype == 1) {
+		      /* XML body of the file is starting here */
+		      inxmlheader = 0;
+		      nestinglevel = 1;
+		      argdefault[0] = '\0';
+		      /* Remove the whole header of the XML file */
+		      if (debug) 
+			fprintf(stderr,
+				"    Removing XML file header\n");
+		      memmove(*data, lasttag, 
+			      *data + datalength + 1 - lasttag);
+		      datalength -= lasttag - *data;
+		      scan -= lasttag - *data;
+		      tagwordstart -= lasttag - *data;
+		      lasttag = *data;
+		      lasttagend = NULL;
+		    }
 		  }
 		} else if (operation == 3) { /* Driver XML file (Overview) */
 		  if (strcmp(currtagname, "printer") == 0) {
@@ -568,6 +633,22 @@ parse(char **data, /* I/O - Data to process */
 		    if (tagtype == 1) lastcomments = (char*)lasttagend + 1;
 		  } else if (strcmp(currtagname, "driver") == 0) {
 		    indriver = nestinglevel + 1;
+		    if (tagtype == 1) {
+		      /* XML body of the file is starting here */
+		      inxmlheader = 0;
+		      nestinglevel = 1;
+		      /* Remove the whole header of the XML file */
+		      if (debug) 
+			fprintf(stderr,
+				"    Removing XML file header\n");
+		      memmove(*data, lasttag, 
+			      *data + datalength + 1 - lasttag);
+		      datalength -= lasttag - *data;
+		      scan -= lasttag - *data;
+		      tagwordstart -= lasttag - *data;
+		      lasttag = *data;
+		      lasttagend = NULL;
+		    }
 		  } 
 		} else if (operation == 4) { /* Printer XML file (Overview)*/
 		  if (debug)
@@ -609,7 +690,7 @@ parse(char **data, /* I/O - Data to process */
 			if (debug)
 			  fprintf(stderr, 
 				  "    Resetting Driver/PPD.\n");
-			strcpy(cid, "Postscript");
+			cid[0] = '\0';
 			cppd[0] = '\0';
 		      }
 		    }
@@ -619,6 +700,20 @@ parse(char **data, /* I/O - Data to process */
 		  } else if (strcmp(currtagname, "printer") == 0) {
 		    inprinter = nestinglevel + 1;
 		    if (tagtype == 1) {
+		      /* XML body of the file is starting here */
+		      inxmlheader = 0;
+		      nestinglevel = 1;
+		      /* Remove the whole header of the XML file */
+		      if (debug) 
+			fprintf(stderr,
+				"    Removing XML file header\n");
+		      memmove(*data, lasttag, 
+			      *data + datalength + 1 - lasttag);
+		      datalength -= lasttag - *data;
+		      scan -= lasttag - *data;
+		      tagwordstart -= lasttag - *data;
+		      lasttag = *data;
+		      lasttagend = NULL;
 		      if (debug) fprintf(stderr, 
 					 "    Initializing PPD list.\n");
 		      while(ppdlist != NULL) {
@@ -732,7 +827,7 @@ parse(char **data, /* I/O - Data to process */
 	      }
 	    } else {
 	      intag = 0;
-	      if (tagnamefound == 0) {
+	      if (!inxmlheader && (tagnamefound == 0)) {
 		fprintf(stderr, "XML error: Tag without name %s, line %d!\n",
 			filename, linecount);
 		exit(1);
@@ -777,18 +872,47 @@ parse(char **data, /* I/O - Data to process */
 			   <drivers> section of printer XML file */
 			if (debug)
 			  fprintf(stderr,
-				  "      Printer XML: Printer/Driver combo confirmed!\n");
+				  "      Printer XML: Printer/Driver combo confirmed by <drivers> section!\n");
 			comboconfirmed = 1;
 		      }
 		    }
 		  }
 		}
+		if (cppd[0]) strcpy(cid, "Postscript");
 		if (nestinglevel < inid) {
 		  inid = 0;
 		  strcpy(cid, currtagbody);
 		  if (debug) fprintf(stderr, 
 				     "    Printer XML: Driver ID: %s\n", cid);
 		}
+		if (nestinglevel < inlang) inlang = 0;
+		if (nestinglevel < inpostscript) {
+		  inpostscript = 0;
+		  if (inlang) {
+		    if (debug) fprintf(stderr, 
+				       "    Printer/Driver/PPD: %s %s %s\n",
+				       cprinter, cid, cppd);
+		  }
+		  if (!strcmp(cid, driver)) {
+		    /* Printer/driver combo already confirmed by
+		       <postscript> section of printer XML file */
+		    if (debug)
+		      fprintf(stderr,
+			      "      Printer XML: Printer/Driver combo confirmed by <postscript> section!\n");
+		    comboconfirmed = 1;
+		  }
+		}
+		if (nestinglevel < inppd) {
+		  inppd = 0;
+		  /* Get the PPD file location without leading 
+		     white space, is empty on empty PPD file location*/
+		  for (s = currtagbody;
+		       (*s != '\0') && (strchr(" \n\r\t", *s) != NULL);
+		       s ++);
+		  strcpy(cppd, s);
+		  if (debug) fprintf(stderr, 
+				     "    PPD URL: %s\n", cppd);
+		}
 	      } else if (operation == 1) { /* Driver XML file */
 		if (nestinglevel < inexecution) inexecution = 0;
 		if (nestinglevel < innopjl) innopjl = 0;
@@ -806,11 +930,13 @@ parse(char **data, /* I/O - Data to process */
 		    if (debug) 
 		      fprintf(stderr, "    Inserting saved printer\n");
 		    l = strlen(printerentry);
-		    memmove(lastprinters + l, lastprinters, 
-			    *data + datalength - lastprinters + 1);
-		    memmove(lastprinters, printerentry, l);
-		    datalength += l;
-		    scan += l;
+		    if (l != 0) {
+		      memmove(lastprinters + l, lastprinters, 
+			      *data + datalength - lastprinters + 1);
+		      memmove(lastprinters, printerentry, l);
+		      datalength += l;
+		      scan += l;
+		    }
 		  }
 		}
 		if (nestinglevel < inprinter) {
@@ -1617,7 +1743,12 @@ parse(char **data, /* I/O - Data to process */
 		}
 		if (nestinglevel < inppd) {
 		  inppd = 0;
-		  strcpy(cppd, currtagbody);
+		  /* Get the PPD file location without leading 
+		     white space, is empty on empty PPD file location*/
+		  for (s = currtagbody;
+		       (*s != '\0') && (strchr(" \n\r\t", *s) != NULL);
+		       s ++);
+		  strcpy(cppd, s);
 		  if (debug) fprintf(stderr, 
 				     "    PPD URL: %s\n", cppd);
 		}
@@ -1678,6 +1809,7 @@ parse(char **data, /* I/O - Data to process */
 	      filename);
       exit(1);
     }
+    if (debug) fprintf(stderr, "    Driver in printer's driver list: %d\n", comboconfirmed); 
   } else if (operation == 1) { /* Driver XML file */
     if (debug) 
       fprintf(stderr,
@@ -1687,6 +1819,7 @@ parse(char **data, /* I/O - Data to process */
 				     driver */
       comboconfirmed = 1;
     }
+    if (debug) fprintf(stderr, "    Printer in driver's printer list: %d\n", comboconfirmed); 
   } else if (operation == 4) { /* Printer XML file (Overview) */
     /* Remove the printer input data */
     **data = '\0';
@@ -1721,9 +1854,12 @@ parse(char **data, /* I/O - Data to process */
 	strcat((char *)(*data), cautodetectentry);
       }
       plistpointer = *printerlist;
+      plistpreventry = NULL;
       while ((plistpointer) &&
-             (strcmp(plistpointer->id, cprinter) != 0))
+             (strcmp(plistpointer->id, cprinter) != 0)) {
+	plistpreventry = plistpointer;
 	plistpointer = (printerlist_t *)(plistpointer->next);
+      }
       if (plistpointer) {
 	strcat((char *)(*data), "\n    <drivers>\n");
 	dlistpointer = plistpointer->drivers;
@@ -1740,7 +1876,8 @@ parse(char **data, /* I/O - Data to process */
 	  strcat((char *)(*data), "    <driverfunctionalityexceptions>\n");
 	  dlistpointer = plistpointer->drivers;
 	  while (dlistpointer) {
-	    if (dlistpointer->functionality != NULL) {
+	    if ((dlistpointer->name != NULL) &&
+		(dlistpointer->functionality != NULL)) {
 	      strcat((char *)(*data),
 		     "      <driverfunctionalityexception>\n");
 	      strcat((char *)(*data), "        <driver>");
@@ -1754,6 +1891,22 @@ parse(char **data, /* I/O - Data to process */
 	  }
 	  strcat((char *)(*data), "    </driverfunctionalityexceptions>\n");
 	}
+	/* We remove every printer entry in the list for which we have found
+	   a printer XML file in the database, so all remaining entries are
+	   of printers which are only mentioned in a driver's printer list
+	   but do not have an XML file in the database. We will treat these
+	   printers later */
+	dlistpointer = plistpointer->drivers;
+	while (dlistpointer) {
+	  dlistpreventry = dlistpointer;
+	  dlistpointer = (driverlist_t *)(dlistpointer->next);
+	  free(dlistpreventry);
+	}
+	if (plistpreventry == NULL)
+	  *printerlist = (printerlist_t *)(plistpointer->next);
+	else
+	  plistpreventry->next = plistpointer->next;
+	free(plistpointer);
       }
       if (ppdlist != NULL) {
 	strcat((char *)(*data), "    <ppds>\n");
@@ -1794,6 +1947,7 @@ main(int  argc,     /* I - Number of command-line arguments */
 {
   int		i,j;		/* Looping vars */
   const char    *tmpstr1;       /* Temporary string constant */
+  char          *t;
 
   const char    *pid = NULL,
                 *driver = NULL,
@@ -1825,6 +1979,7 @@ main(int  argc,     /* I - Number of command-line arguments */
   int           debug2 = 0;
   int           comboconfirmed = 0;
   int           comboconfirmed2 = 0;
+  int           exceptionfound = 0;
   DIR           *optiondir;
   DIR           *driverdir;
   DIR           *printerdir;
@@ -1833,6 +1988,8 @@ main(int  argc,     /* I - Number of command-line arguments */
   printerlist_t *plistpointer;  /* pointers to navigate through the 
 				   printer */
   driverlist_t  *dlistpointer;  /* list for the overview */
+  printerlist_t *plistpreventry;
+  driverlist_t  *dlistpreventry;
   idlist_t      *idlist;        /* I - ID translation table */
   
   /* Show the help message whem no command line arguments are given */
@@ -1982,10 +2139,27 @@ main(int  argc,     /* I - Number of command-line arguments */
 	      libdir, pid);
       printerbuffer = loadfile(printerfilename);
       if (printerbuffer == NULL) {
-	fprintf(stderr, 
-		"Printer file %s corrupted, missing, or not readable!\n",
-		printerfilename);
-	exit(1);
+	printerbuffer = malloc(1024);
+	make = strdup(pid);
+	model = strchr(make, '-');
+	if (model) {
+	  t = (char *)model;
+	  *t = '\0';
+	  model ++;
+	} else { 
+	  model = "Unknown model";
+	}
+	t = (char *)make;
+	while (*t) {
+	  if (*t == '_') *t = ' ';
+	  t ++;
+	}
+	t = (char *)model;
+	while (*t) {
+	  if (*t == '_') *t = ' ';
+	  t ++;
+	}
+	sprintf((char *)printerbuffer, "<printer id=\"printer/%s\">\n <make>%s</make>\n <model>%s</model>\n <noxmlentry />\n<printer>\n", pid, make, model);
       } else {
 	fprintf(stderr, 
 		"WARNING: Obsolete printer ID used, using %s instead!\n",
@@ -2009,7 +2183,7 @@ main(int  argc,     /* I - Number of command-line arguments */
 		driverfilename);
 	exit(1);
       } else {
-	driverbuffer = malloc(1024);
+	driverbuffer = malloc(4096);
 	sprintf((char *)driverbuffer, "<driver id=\"driver/%s\">\n <name>%s</name>\n <url></url>\n <execution>\n  <filter />\n  <prototype></prototype>\n </execution>\n <printers>\n  <printer>\n   <id>printer/%s</id>\n  </printer>\n </printers>\n</driver>", driver, driver, pid);
       }
     } else {
@@ -2019,13 +2193,15 @@ main(int  argc,     /* I - Number of command-line arguments */
 	      (const char **)defaultsettings, num_defaultsettings, &nopjl,
 	      idlist, debug2);
       if ((!comboconfirmed) && (!comboconfirmed2)) {
-	fprintf(stderr, "The printer %s %s (ID: %s) is not supported by the driver %s!\n",
-		make, model, pid, driver);
+	fprintf(stderr, "The printer %s is not supported by the driver %s!\n",
+		pid, driver);
 	exit(1);
       }
       if (debug) {
 	if (nopjl) {
 	  fprintf(stderr, "  Driver forbids PJL options!\n");
+	} else {
+	  fprintf(stderr, "  Driver allows PJL options!\n");
 	}
       }
     
@@ -2079,7 +2255,7 @@ main(int  argc,     /* I - Number of command-line arguments */
     
     /* Output the result on STDOUT */
     if (debug) fprintf(stderr, "Putting out result!\n");
-    printf("<foomatic>%s%s\n<options>\n", printerbuffer, driverbuffer);
+    printf("<foomatic>\n%s%s\n<options>\n", printerbuffer, driverbuffer);
     for (i = 0; i < num_optbuffers; i++) {
       printf("%s", optbuffers[i]);
     }
@@ -2214,10 +2390,107 @@ main(int  argc,     /* I - Number of command-line arguments */
       }
     }
 
-    printf("</overview>\n");
-
     closedir(printerdir);
 
+    if (debug) {
+      plistpointer = printerlist;
+      while (plistpointer) {
+	fprintf(stderr, "Printer: %s\n", plistpointer->id);
+	dlistpointer = plistpointer->drivers;
+	while (dlistpointer) {
+	  fprintf(stderr, "   Driver: %s\n", dlistpointer->name);
+	  if (dlistpointer->functionality != NULL)
+	    fprintf(stderr, "    %s\n", dlistpointer->functionality);
+	  dlistpointer = (driverlist_t *)(dlistpointer->next);
+	}
+	plistpointer = (printerlist_t *)(plistpointer->next);
+      }
+    }
+
+    if (overview < 2) {
+      /* Now show all printers which are only mentioned in the lists of
+	 supported prnters of the drivers and which not have a Foomatic
+	 printer XML entry. This we do only for a general overview, not
+         for an overview of valid PPD files ("-C" option) */
+      while (printerlist) {
+	if (printerlist->id && strcmp(printerlist->id, "noproto")) {
+	  if (debug) fprintf(stderr, "    Printer only mentioned in driver XML files:\n      Printer ID: |%s|\n",
+			     printerlist->id);
+	  /*strcpy(printerlist->id, translateid(printerlist->id, idlist));*/
+	  printf("  <printer>\n    <id>");
+	  printf("%s", printerlist->id);
+	  make = printerlist->id;
+	  model = strchr(make, '-');
+	  if (model) {
+	    t = (char *)model;
+	    *t = '\0';
+	    model ++;
+	  } else { 
+	    model = "Unknown model";
+	  }
+	  t = (char *)make;
+	  while (*t) {
+	    if (*t == '_') *t = ' ';
+	    t ++;
+	  }
+	  t = (char *)model;
+	  while (*t) {
+	    if (*t == '_') *t = ' ';
+	    t ++;
+	  }
+	  printf("</id>\n    <make>");
+	  printf("%s", make);
+	  printf("</make>\n    <model>");
+	  printf("%s", model);
+	  printf("</model>\n    <noxmlentry />\n");
+	  dlistpointer = printerlist->drivers;
+	  exceptionfound = 0;
+	  if (dlistpointer) {
+	    printf("    <drivers>\n");
+	    while (dlistpointer) {
+	      if (dlistpointer->name) {
+		printf("      <driver>");
+		printf("%s", dlistpointer->name);
+		printf("</driver>\n");
+		if (dlistpointer->functionality != NULL) exceptionfound = 1;
+	      }
+	      dlistpointer = (driverlist_t *)(dlistpointer->next);
+	    }
+	    printf("    </drivers>\n");
+	  }
+	  if (exceptionfound) {
+	    printf("    <driverfunctionalityexceptions>\n");
+	    dlistpointer = printerlist->drivers;
+	    while (dlistpointer) {
+	      if ((dlistpointer->functionality != NULL) &&
+		  (dlistpointer->name != NULL)) {
+		printf("      <driverfunctionalityexception>\n");
+		printf("        <driver>");
+		printf("%s", dlistpointer->name);
+		printf("</driver>\n");
+		printf("%s", dlistpointer->functionality);
+		printf("\n      </driverfunctionalityexception>\n");
+	      }
+	      dlistpointer = (driverlist_t *)(dlistpointer->next);
+	    }
+	    printf("    </driverfunctionalityexceptions>\n");
+	  }
+	  printf("  </printer>\n");
+	}
+	dlistpointer = printerlist->drivers;
+	while (dlistpointer) {
+	  dlistpreventry = dlistpointer;
+	  dlistpointer = (driverlist_t *)(dlistpointer->next);
+	  free(dlistpreventry);
+	}
+	plistpreventry = printerlist;
+	printerlist = (printerlist_t *)(printerlist->next);
+	free(plistpreventry);
+      }
+    }
+
+    printf("</overview>\n");
+      
   }
     
   /* Done */
